@@ -1,21 +1,16 @@
-from typing import Tuple, Callable, Union, TypeVar
+from typing import Tuple, Dict, Callable, Union, TypeVar, cast, Any
 import inspect
 from dataclasses import is_dataclass, fields
 
+from libbiomedit.lib.secret import Secret, reveal
+from libbiomedit.lib.classify import is_secret
 
-class Secret(str):
-    """A class to mark arguments as sensitive"""
-
-    def __repr__(self):
-        return "***"
-
-    def __format__(self, format_spec):
-        return str.__format__("***", format_spec)
+__all__ = ["Secret", "reveal"]
 
 
 def enforce_secret_by_signature(
-    f: Callable, args: tuple, kwargs: dict
-) -> Tuple[tuple, dict]:
+    f: Callable[..., Any], args: Tuple[Any, ...], kwargs: Dict[str, Any]
+) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
     sig = inspect.signature(f)
     sanitized_kwargs = {
         key: enforce_secret(val, sig.parameters[key].annotation)
@@ -31,13 +26,13 @@ def enforce_secret_by_signature(
 T = TypeVar("T")
 
 
-def enforce_secret(val: T, t: type) -> Union[T, Secret]:
-    if t is Secret or (  # check for Optional / Union:
+def enforce_secret(val: T, t: type) -> Union[T, Secret[str]]:
+    if is_secret(t) or (  # check for Optional / Union:
         getattr(t, "__origin__", None) is Union
-        and Secret in getattr(t, "__args__", ())
+        and any(is_secret(t) for t in getattr(t, "__args__", ()))
         and isinstance(val, str)
     ):
-        return Secret(val)
+        return Secret(cast(str, val))
     if is_dataclass(val) and issubclass(type(val), t):
         t = type(val)
     if is_dataclass(t):
