@@ -1,7 +1,8 @@
 #!/usr/bin/env python3
-"""Secure Encryption and Transfer Tool GUI"""
+"""Secure Encryption and Transfer Tool GUI."""
 
 import sys
+from typing import Any
 
 from . import main_window
 from .pyside import QtWidgets, open_window
@@ -12,7 +13,7 @@ from .resources import rc_icons  # pylint: disable=unused-import
 from ..utils.log import log_to_rotating_file
 
 
-def run():
+def run() -> Any:
     application = QtWidgets.QApplication(sys.argv)
     window = main_window.MainWindow()
     log_to_rotating_file(
