@@ -60,7 +60,7 @@ class DecryptTab(TabMixin):
         )
 
         output_location = PathInput(path=self.app_data.encrypt_output_location)
-        output_location.status_tip = "Destination folder for the decrypted packages"
+        output_location.setStatusTip("Destination folder for the decrypted packages")
         output_location.on_path_change(
             partial(setattr, self.app_data, "decrypt_output_location")
         )
