@@ -1,6 +1,14 @@
 import html
 from pathlib import Path
-from typing import Iterator, Optional, Sequence, Callable
+from typing import (
+    Iterator,
+    Optional,
+    Sequence,
+    Tuple,
+    Callable,
+    Any,
+    cast,
+)
 
 from sett.core.error import UserError
 from .component import (
@@ -9,14 +17,15 @@ from .component import (
     ToolBar,
     LineEdit,
     warning_callback,
+    show_warning,
     grid_layout,
     GridLayoutCell,
 )
 from .keys_refresh import load_authority_key_threaded
 from .parallel import run_thread
 from .pyside import QtCore, QtGui, QtWidgets, QAction, open_window
-from ..core import crypt
-from ..core.crypt import gpg
+from ..core import crypt, gpg
+from ..core.secret import Secret
 from ..utils.config import Config
 from ..workflows import (
     upload_keys as upload_keys_workflow,
@@ -25,7 +34,7 @@ from ..workflows import (
 
 
 class KeysTab(QtWidgets.QWidget):
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any):
         super().__init__(*args, **kwargs)
         self.app_data = self.parent().app_data
 
@@ -77,13 +86,13 @@ class KeysTab(QtWidgets.QWidget):
             "Refresh keys from the local keyring",
             self,
         )
-        action_refresh_keys.triggered.connect(
-            lambda: (
-                self.update_private_keys(),
-                self.update_public_keys(),
-                self.update_display_selected_pub_key(),
-            )
-        )
+
+        def refresh_keys() -> None:
+            self.update_private_keys()
+            self.update_public_keys()
+            self.update_display_selected_pub_key()
+
+        action_refresh_keys.triggered.connect(refresh_keys)
 
         toolbar = ToolBar("Key management", self)
         toolbar.addAction(action_generate_key)
@@ -172,13 +181,14 @@ class KeysTab(QtWidgets.QWidget):
     def create_public_keys_actions(self) -> Iterator[QAction]:
         selection_model = self.pub_keys_view.selectionModel()
 
-        def offline_action(icon: str, tip: str, selection: bool = True):
-            """Force disable button in offline mode"""
+        def offline_action(icon: str, tip: str, selection: bool = True) -> QAction:
+            """Force disable button in offline mode."""
+
             icon_obj = f":icon/feather/{icon}.png"
             if self.app_data.config.offline:
                 action = QAction(
                     QtGui.QIcon(icon_obj),
-                    f"{tip} (not available in the offline mode)",
+                    f"{tip} (not available in offline mode)",
                     self,
                 )
                 action.setEnabled(False)
@@ -194,7 +204,12 @@ class KeysTab(QtWidgets.QWidget):
                     action = QAction(QtGui.QIcon(icon_obj), tip, self)
             return action
 
-        for action, fn in (
+        # Create a dict of actions (functions) to associate to each
+        # action button of the GUI.
+        # The reason to create a dict object before looping over it is so that
+        # mypy gets the correct typing information (for some reason
+        # SelectionAction is not recognized as a subtype of QAction).
+        functions_by_action: Sequence[Tuple[QAction, Callable[..., Any]]] = (
             (
                 offline_action(
                     "download-cloud",
@@ -232,12 +247,14 @@ class KeysTab(QtWidgets.QWidget):
                 ),
                 self.import_key,
             ),
-        ):
+        )
+        for action, fn in functions_by_action:
             action.triggered.connect(fn)
             yield action
 
-    def update_private_keys(self):
-        """Retrieve all private keys from the user's local keyring"""
+    def update_private_keys(self) -> None:
+        """Retrieve/reload all private keys from the user's local keyring and
+        store them in the application's memory."""
         keys_private = self.app_data.config.gpg_store.list_sec_keys()
         try:
             default_key = (
@@ -256,25 +273,34 @@ class KeysTab(QtWidgets.QWidget):
             {f"{key.uids[0]} {key.key_id}": key for key in keys_private}
         )
 
-    def update_public_keys(self):
-        """Retrieve all public keys from the user's local keyring"""
+    def update_public_keys(self) -> None:
+        """Retrieve/reload all public keys from the user's local keyring and
+        store them in the application's memory."""
         keys_public = self.app_data.config.gpg_store.list_pub_keys(sigs=True)
         self.app_data.pub_keys_model.set_data(
             {f"{key.uids[0]} {key.key_id}": key for key in keys_public}
         )
 
-    def update_keys(self):
-        """Update/refresh selected keys from the keyserver"""
+    def get_selected_keys(self) -> Tuple[gpg.Key, ...]:
+        """Returns the gpg.Key objects corresponding to the keys currently
+        selected in the GUI.
+        """
+        # NOTE: cast() is needed here as selectedIndexes() wrongly hints that
+        #       it returns List[int] when in fact it returns
+        #       Sequence[QtCore.QModelIndex].
+        selected_indexes = (
+            cast(QtCore.QModelIndex, x) for x in self.pub_keys_view.selectedIndexes()
+        )
+        selected_keys = (index.model().get_value(index) for index in selected_indexes)
+        return tuple(selected_keys)
+
+    def update_keys(self) -> None:
+        """Update/refresh selected keys from the keyserver."""
         show_ok = ok_message("Updated keys", "Keys have been successfully updated.")
-        show_warning = warning_callback("GPG key update error")
+        keys_to_update = self.get_selected_keys()
+        if keys_to_update:
 
-        selected_keys = [
-            index.model().get_value(index).fingerprint
-            for index in self.pub_keys_view.selectedIndexes()
-        ]
-        if selected_keys:
-
-            def on_result():
+            def on_result() -> None:
                 self.update_public_keys()
                 self.update_display_selected_pub_key()
                 show_ok()
@@ -282,17 +308,17 @@ class KeysTab(QtWidgets.QWidget):
             run_thread(
                 crypt.download_keys,
                 f_kwargs=dict(
-                    fingerprints=selected_keys,
+                    fingerprints=[key.fingerprint for key in keys_to_update],
                     keyserver=self.app_data.config.keyserver_url,
                     gpg_store=self.app_data.config.gpg_store,
                 ),
                 report_config=self.app_data.config,
-                forward_errors=show_warning,
+                forward_errors=warning_callback("GPG key update error"),
                 signals=dict(result=on_result),
             )
 
-    def import_key(self):
-        """Import a GPG key from a local file"""
+    def import_key(self) -> None:
+        """Import a GPG key from a local file."""
         path = QtWidgets.QFileDialog.getOpenFileName(
             self, "Select GPG key file", str(Path.home())
         )[0]
@@ -313,7 +339,10 @@ class KeysTab(QtWidgets.QWidget):
             msg.setText(format(e))
             open_window(msg)
 
-    def delete_keys(self):
+    def delete_keys(self) -> None:
+        """Delete the selected public keys from the user's local keyring. Only
+        public keys with no associated private key can be deleted.
+        """
         msg = QtWidgets.QMessageBox()
         msg.setIcon(QtWidgets.QMessageBox.Question)
         msg.setWindowTitle("Delete public key")
@@ -324,11 +353,8 @@ class KeysTab(QtWidgets.QWidget):
         msg_warn.setIcon(QtWidgets.QMessageBox.Warning)
         priv_keys = self.app_data.config.gpg_store.list_sec_keys()
 
-        selected_keys = [
-            index.model().get_value(index)
-            for index in self.pub_keys_view.selectedIndexes()
-        ]
-        for key in selected_keys:
+        keys_to_delete = self.get_selected_keys()
+        for key in keys_to_delete:
             if any(k for k in priv_keys if key.fingerprint == k.fingerprint):
                 msg_warn.setText(
                     "Unable to delete key:\n\n"
@@ -341,7 +367,7 @@ class KeysTab(QtWidgets.QWidget):
                 open_window(msg_warn)
                 continue
             msg.setDetailedText(self.key_to_text(key))
-            if key is selected_keys[0]:
+            if key is keys_to_delete[0]:
                 click_show_details(msgbox=msg)
             status = open_window(msg)
             if status == QtWidgets.QMessageBox.Ok:
@@ -356,7 +382,10 @@ class KeysTab(QtWidgets.QWidget):
                 self.text_panel.clear()
         self.update_public_keys()
 
-    def upload_key(self):
+    def upload_key(self) -> None:
+        """Upload the selected keys to the keyserver specific in the user's
+        config file.
+        """
         msg = QtWidgets.QMessageBox()
         msg.setWindowTitle("Send public key")
         msg.setText("Do you want to upload the selected key to the key server?")
@@ -365,15 +394,11 @@ class KeysTab(QtWidgets.QWidget):
         show_ok = ok_message(
             "Send public key", "Key has been successfully uploaded to the keyserver."
         )
-        show_warning = warning_callback("GPG key upload error")
 
-        selected_keys = [
-            index.model().get_value(index)
-            for index in self.pub_keys_view.selectedIndexes()
-        ]
-        for key in selected_keys:
+        keys_to_upload = self.get_selected_keys()
+        for key in keys_to_upload:
             msg.setDetailedText(self.key_to_text(key))
-            if key is selected_keys[0]:
+            if key is keys_to_upload[0]:
                 click_show_details(msgbox=msg)
             status = open_window(msg)
             if status == QtWidgets.QMessageBox.Ok:
@@ -384,19 +409,16 @@ class KeysTab(QtWidgets.QWidget):
                     ),
                     report_config=self.app_data.config,
                     capture_loggers=(upload_keys_workflow.logger,),
-                    forward_errors=show_warning,
+                    forward_errors=warning_callback("GPG key upload error"),
                     signals=dict(result=lambda _: show_ok()),
                 )
 
-    def send_signature_request(self):
-        selected_keys = [
-            index.model().get_value(index)
-            for index in self.pub_keys_view.selectedIndexes()
-        ]
-        request_signature(selected_keys, self.app_data.config, self)
+    def send_signature_request(self) -> None:
+        """Send a request to the key authority to sign the selected keys."""
+        request_signature(self.get_selected_keys(), self.app_data.config, self)
 
-    def _update_display(self, index: QtCore.QModelIndex):
-        """Display key info summary in GUI text panel"""
+    def _update_display(self, index: QtCore.QModelIndex) -> None:
+        """Display key info summary in GUI text panel."""
         style = (
             "<style>"
             "th {text-align: left; padding: 0 20px 5px 0;}"
@@ -412,16 +434,16 @@ class KeysTab(QtWidgets.QWidget):
             except IndexError:
                 self.text_panel.setHtml("")
 
-    def update_display_selected_pub_key(self):
-        """Refresh the displayed key info of the currently selected public key(s)"""
+    def update_display_selected_pub_key(self) -> None:
+        """Refresh the displayed key info of the currently selected public keys."""
         self._update_display(self.pub_keys_view.selectionModel().currentIndex())
 
 
 class KeyDownloadDialog(QtWidgets.QDialog):
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any):
         super().__init__(*args, **kwargs)
         self.setWindowTitle("Download public keys from keyserver")
-        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)
+        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)  # type: ignore
 
         self.search_string = LineEdit()
         self.btn_search = QtWidgets.QPushButton("Search")
@@ -464,16 +486,16 @@ class KeyDownloadDialog(QtWidgets.QDialog):
             )
         )
 
-    def set_number_of_keys_found(self, n: Optional[int]):
+    def set_number_of_keys_found(self, n: Optional[int]) -> None:
         self.label_results.setText(f"Number of keys found: {'-' if n is None else n}")
 
-    def search_keys(self):
+    def search_keys(self) -> None:
         self.btn_search.setEnabled(False)
         self.key_list_view.selectionModel().clear()
         self.key_list_view.model().setStringList([])
         self.set_number_of_keys_found(None)
 
-        def on_result(keys):
+        def on_result(keys: Sequence[gpg.KeyInfo]) -> None:
             self.key_list_view.model().setStringList(
                 [f"{k.uid} {k.fingerprint}" for k in keys]
             )
@@ -490,42 +512,45 @@ class KeyDownloadDialog(QtWidgets.QDialog):
             signals=dict(result=on_result),
         )
 
-    def download_selected(self):
-        """Download keys selected in the key-search pop-up to local keyring"""
-        key_ids = [
-            index.model().data(index).split()[-1]
-            for index in self.key_list_view.selectedIndexes()
-        ]
-        if not key_ids:
-            return
-
-        def on_result():
-            self.parent().update_public_keys()
-            self.parent().update_display_selected_pub_key()
-
-        self.btn_download.setEnabled(False)
-        show_warning = warning_callback("GPG key search error")
-        run_thread(
-            crypt.download_keys,
-            f_kwargs=dict(
-                fingerprints=key_ids,
-                keyserver=self.parent().app_data.config.keyserver_url,
-                gpg_store=self.parent().app_data.config.gpg_store,
-            ),
-            report_config=self.parent().app_data.config,
-            forward_errors=show_warning,
-            signals=dict(
-                result=on_result,
-                finished=lambda: self.btn_download.setEnabled(True),
-            ),
+    def download_selected(self) -> None:
+        """Download keys selected in the key-search pop-up to the user's
+        local keyring.
+        """
+        # NOTE: cast() is needed here as selectedIndexes() wrongly hints that
+        #       it returns List[int] when in fact it returns
+        #       Sequence[QtCore.QModelIndex].
+        selected_indexes = (
+            cast(QtCore.QModelIndex, x) for x in self.key_list_view.selectedIndexes()
         )
+        keys_to_download = [index.model().data(index) for index in selected_indexes]
+        if keys_to_download:
+
+            def on_result() -> None:
+                self.parent().update_public_keys()
+                self.parent().update_display_selected_pub_key()
+
+            self.btn_download.setEnabled(False)
+            run_thread(
+                crypt.download_keys,
+                f_kwargs=dict(
+                    fingerprints=[x.split()[-1] for x in keys_to_download],
+                    keyserver=self.parent().app_data.config.keyserver_url,
+                    gpg_store=self.parent().app_data.config.gpg_store,
+                ),
+                report_config=self.parent().app_data.config,
+                forward_errors=warning_callback("GPG key search error"),
+                signals=dict(
+                    result=on_result,
+                    finished=lambda: self.btn_download.setEnabled(True),
+                ),
+            )
 
 
 class KeyGenDialog(QtWidgets.QDialog):
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any):
         super().__init__(*args, **kwargs)
         self.setWindowTitle("Generate new key pair")
-        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)
+        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowContextHelpButtonHint)  # type: ignore
 
         self.text_name_full = LineEdit()
         self.text_name_extra = LineEdit()
@@ -570,19 +595,19 @@ class KeyGenDialog(QtWidgets.QDialog):
             )
         )
 
-    def toggle_password_visibility(self, show: bool):
+    def toggle_password_visibility(self, show: bool) -> None:
         mode = QtWidgets.QLineEdit.Normal if show else QtWidgets.QLineEdit.Password
         self.text_pass.setEchoMode(mode)
         self.text_pass_repeat.setEchoMode(mode)
 
-    def clear_form(self):
+    def clear_form(self) -> None:
         self.text_name_full.clear()
         self.text_name_extra.clear()
         self.text_email.clear()
         self.text_pass.clear()
         self.text_pass_repeat.clear()
 
-    def post_key_creation(self, key: gpg.Key):
+    def post_key_creation(self, key: gpg.Key) -> None:
         msg = QtWidgets.QMessageBox(self)
         msg.setWindowTitle("GPG Key Generation")
         msg.setIcon(QtWidgets.QMessageBox.Information)
@@ -596,7 +621,7 @@ class KeyGenDialog(QtWidgets.QDialog):
                 "Your new key has been successfully generated.\n\n"
                 "Additionally, a revocation certificate was also created. "
                 "It can be used to revoke your key in the eventuality that it "
-                "gets compromised, lost, or that you forgot your password.\n"
+                "gets compromised, lost, or that you forget your password.\n"
                 "Please store the revocation certificate below in a safe "
                 "location, as anyone can use it to revoke your key."
             )
@@ -608,10 +633,9 @@ class KeyGenDialog(QtWidgets.QDialog):
         except UserError:
             msg.setIcon(QtWidgets.QMessageBox.Warning)
             msg.setText(
-                "Key has been successfully generated. "
-                "However, it was not possible to create a revocation "
-                "certificate. "
-                "Execute the following command to create the certificate\n\n"
+                "Key has been successfully generated. However, it was not "
+                "possible to create a revocation certificate for it. Please "
+                "execute the following command to create the certificate:\n\n"
                 f"gpg --gen-revoke {key.fingerprint}"
             )
         finally:
@@ -626,8 +650,7 @@ class KeyGenDialog(QtWidgets.QDialog):
             self.parent().update_public_keys()
             self.close()
 
-    def create_private_key(self):
-        show_warning = warning_callback("GPG Key Generation Error")
+    def create_private_key(self) -> None:
         self.btn_run.setEnabled(False)
         name_full = self.text_name_full.text().strip()
         name_extra = self.text_name_extra.text().strip()
@@ -635,17 +658,24 @@ class KeyGenDialog(QtWidgets.QDialog):
             if not name_extra.startswith("(") and not name_extra.endswith(")"):
                 name_extra = f"({name_extra})"
             name_full = name_full + " " + name_extra
+        if self.text_pass.text() != self.text_pass_repeat.text():
+            show_warning(
+                "GPG Key Generation Error",
+                "Password and repeated password do not match.",
+                self,
+            )
+            self.btn_run.setEnabled(True)
+            return
 
         run_thread(
             crypt.create_key,
             f_kwargs=dict(
                 full_name=name_full,
                 email=self.text_email.text(),
-                pwd=self.text_pass.text(),
-                pwd_repeat=self.text_pass_repeat.text(),
+                pwd=Secret(self.text_pass.text()),
                 gpg_store=self.parent().app_data.config.gpg_store,
             ),
-            forward_errors=show_warning,
+            forward_errors=warning_callback("GPG Key Generation Error"),
             report_config=self.parent().app_data.config,
             signals=dict(
                 result=self.post_key_creation,
@@ -654,7 +684,7 @@ class KeyGenDialog(QtWidgets.QDialog):
         )
 
 
-def click_show_details(msgbox: QtWidgets.QMessageBox):
+def click_show_details(msgbox: QtWidgets.QMessageBox) -> None:
     for button in msgbox.buttons():
         if msgbox.buttonRole(button) is QtWidgets.QMessageBox.ButtonRole.ActionRole:
             button.click()
@@ -664,7 +694,7 @@ def request_signature(
     keys: Sequence[gpg.Key],
     config: Config,
     parent: Optional[QtWidgets.QWidget] = None,
-):
+) -> None:
     """Request the specified keys to be signed by the key validation authority.
     This triggers the following:
      * The specified keys are uploaded to the keyserver.
@@ -678,7 +708,6 @@ def request_signature(
     show_ok = ok_message(
         "Key signing request", "Key signing request has been sent.", parent=parent
     )
-    show_warning = warning_callback("Key signing request", parent=parent)
 
     for key in keys:
         msg.setDetailedText(KeysTab.key_to_text(key))
@@ -693,13 +722,15 @@ def request_signature(
                     config=config,
                 ),
                 capture_loggers=(request_sigs_workflow.logger,),
-                forward_errors=show_warning,
+                forward_errors=warning_callback("Key signing request", parent=parent),
                 report_config=config,
                 signals=dict(result=lambda _: show_ok()),
             )
 
 
-def ok_message(title: str, msg: str, parent=None) -> Callable:
+def ok_message(
+    title: str, msg: str, parent: Optional[QtWidgets.QWidget] = None
+) -> Callable[[], Any]:
     msg_ok = QtWidgets.QMessageBox(parent)
     msg_ok.setIcon(QtWidgets.QMessageBox.Information)
     msg_ok.setWindowTitle(title)
