@@ -1,7 +1,9 @@
-# pylint: disable=unused-import
-from typing import Optional, Any, Protocol
+import os
+from typing import Optional, Any
 
 try:
+    if os.environ.get("SETT_LEGACY", "") == "true":
+        raise ImportError("Enforce legacy mode")  # pylint: disable=misplaced-bare-raise
     from PySide6 import QtCore as _QtCore
     from PySide6 import QtGui as _QtGui
     from PySide6 import QtWidgets as _QtWidgets
