@@ -7,7 +7,17 @@ from contextlib import contextmanager
 from copy import copy
 from functools import wraps
 from pathlib import Path
-from typing import Callable, Tuple, Dict, cast, ContextManager, Optional
+from typing import (
+    Callable,
+    Tuple,
+    Dict,
+    cast,
+    ContextManager,
+    Optional,
+    Generator,
+    Any,
+    TypeVar,
+)
 
 import colorama  # type: ignore
 
@@ -29,7 +39,7 @@ class EmojiLevelFormatter(logging.Formatter):
         logging.CRITICAL: "☢️",
     }
 
-    def format(self, record: logging.LogRecord):
+    def format(self, record: logging.LogRecord) -> str:
         emoji = self.log_style.get(record.levelno)
         if emoji:
             # Assigning to `levelname` changes the value also for other
@@ -58,9 +68,9 @@ class ColoredTerminalFormatter(logging.Formatter):
 @contextmanager
 def log_task(
     message: str,
-    ok_logger: Callable = logging.info,
-    err_logger: Callable = logging.error,
-):
+    ok_logger: Callable[..., None] = logging.info,
+    err_logger: Callable[..., None] = logging.error,
+) -> Generator[None, None, None]:
     """Context manager / decorator that logs a user-defined message with level INFO upon
     entering the context manager, and either "OK" or "FAILED" upon exiting
     depending on whether exceptions occurred during its body or not.
@@ -82,12 +92,17 @@ def log_task(
     ok_logger("OK")
 
 
-def log_timing(logger: logging.Logger):
+R = TypeVar("R")
+
+
+def log_timing(
+    logger: logging.Logger,
+) -> Callable[[Callable[..., R]], Callable[..., R]]:
     """A function decorator, logging processing time taken by the contained code"""
 
-    def decorator(f):
+    def decorator(f: Callable[..., R]) -> Callable[..., R]:
         @wraps(f)
-        def wrapped_f(*args, **kwargs):
+        def wrapped_f(*args: Any, **kwargs: Any) -> R:
             tic = time.perf_counter()
             r = f(*args, **kwargs)
             tac = time.perf_counter()
@@ -109,7 +124,9 @@ def sec_to_hours(seconds: int) -> str:
     )
 
 
-def log_runtime_info(logger: logging.Logger):
+def log_runtime_info(
+    logger: logging.Logger,
+) -> Callable[[Callable[..., R]], Callable[..., R]]:
     """Logs version of the application, dependencies, and platform.
 
     - application version
@@ -118,9 +135,9 @@ def log_runtime_info(logger: logging.Logger):
     - version of the operating system where the application is running
     """
 
-    def decorator(f):
+    def decorator(f: Callable[..., R]) -> Callable[..., R]:
         @wraps(f)
-        def wrapped_f(*args, **kwargs):
+        def wrapped_f(*args: Any, **kwargs: Any) -> R:
             logger.info(RUNTIME_INFO)
             return f(*args, **kwargs)
 
@@ -132,7 +149,7 @@ def log_runtime_info(logger: logging.Logger):
 class LoggerExt(logging.Logger):
     """Logger class with additional functionality"""
 
-    def log_task(self, message: str) -> ContextManager:
+    def log_task(self, message: str) -> ContextManager[None]:
         """A context manager to log a task which can potentially fail"""
         return log_task(message, ok_logger=self.info, err_logger=self.error)
 
@@ -143,17 +160,19 @@ logging.setLoggerClass(LoggerExt)
 class StringIoHandler(logging.Handler):
     """Logging handler logging to a StringIo."""
 
-    def __init__(self, *args, **kwargs):
+    def __init__(self, *args: Any, **kwargs: Any):
         self.log = io.StringIO()
         super().__init__(*args, **kwargs)
 
-    def emit(self, record):
-        self.log.write(self.formatter.format(record))
+    def emit(self, record: logging.LogRecord) -> None:
+        self.log.write(self.format(record))
         self.log.write("\n")
 
 
 @contextmanager
-def log_to_memory(logger: Optional[logging.Logger] = None):
+def log_to_memory(
+    logger: Optional[logging.Logger] = None,
+) -> Generator[io.StringIO, None, None]:
     mem_handler = StringIoHandler()
     mem_handler.setFormatter(logging.Formatter(LOG_FORMAT_WITH_TIME_AND_PID))
     if logger is None:
@@ -166,7 +185,7 @@ def log_to_memory(logger: Optional[logging.Logger] = None):
     logger.setLevel(saved_level)
 
 
-def create_logger(name, logger_level: int = logging.DEBUG) -> LoggerExt:
+def create_logger(name: str, logger_level: int = logging.DEBUG) -> LoggerExt:
     """Instantiate and return a new logger object with the user selected name
      and logging level (e.g. DEBUG, INFO).
 
